__copyright__ = "Copyright © Stichting SciPost (SciPost Foundation)"
__license__ = "AGPL v3"


import datetime

from django.test import TestCase

from ..services import ArxivCaller, DOICaller

from submissions.factories import PreprintServerFactory


class ArxivCallerTest(TestCase):
    def setUp(self) -> None:
        # Create the arXiv preprint server
        PreprintServerFactory.arxiv()

    def test_identifier_new_style(self):
        caller = ArxivCaller("1612.07611v1")
        self.assertTrue(caller.is_valid)
        correct_data = {
            "arxiv_link": "https://arxiv.org/abs/1612.07611v1",
            "author_list": "Roman Krčmár, Andrej Gendiar, Tomotoshi Nishino",
            "abstract": "The Berezinskii-Kosterlitz-Thouless (BKT) transitions of the six-state clock model on the square lattice are investigated by means of the corner-transfer matrix renormalization group method. The classical analogue of the entanglement entropy $S( L, T )$ is calculated for $L$ by $L$ square system up to $L = 129$, as a function of temperature $T$. The entropy has a peak at $T = T^{*}_{~}( L )$, where the temperature depends on both $L$ and boundary conditions. Applying the finite-size scaling to $T^{*}_{~}( L )$ and assuming the presence of BKT transitions, the transition temperature is estimated to be $T_1^{~} = 0.70$ and $T_2^{~} = 0.88$. The obtained results agree with previous analyses. It should be noted that no thermodynamic function is used in this study.",
            "pub_abstract": "The Berezinskii-Kosterlitz-Thouless (BKT) transitions of the six-state clock model on the square lattice are investigated by means of the corner-transfer matrix renormalization group method. The classical analogue of the entanglement entropy $S( L, T )$ is calculated for $L$ by $L$ square system up to $L = 129$, as a function of temperature $T$. The entropy has a peak at $T = T^{*}_{~}( L )$, where the temperature depends on both $L$ and boundary conditions. Applying the finite-size scaling to $T^{*}_{~}( L )$ and assuming the presence of BKT transitions, the transition temperature is estimated to be $T_1^{~} = 0.70$ and $T_2^{~} = 0.88$. The obtained results agree with previous analyses. It should be noted that no thermodynamic function is used in this study.",
            "title": "Phase transition of the six-state clock model observed from the entanglement entropy",
            "preprint_link": "https://arxiv.org/abs/1612.07611v1",
            "pub_date": datetime.date(2016, 12, 22),
        }
        del caller.data["preprint_server"]  # Remove server key before comparison
        self.assertDictEqual(caller.data, correct_data)

    def test_identifier_old_style(self):
        caller = ArxivCaller("cond-mat/0612480")
        self.assertTrue(caller.is_valid)
        correct_data = {
            "arxiv_link": "https://arxiv.org/abs/cond-mat/0612480v2",
            "pub_date": datetime.date(2006, 12, 19),
            "author_list": "Kouji Ueda, Chenglong Jin, Naokazu Shibata, Yasuhiro Hieida, Tomotoshi Nishino",
            "abstract": "A kind of least action principle is introduced for the discrete time evolution of one-dimensional quantum lattice models. Based on this principle, we obtain an optimal condition for the matrix product states on succeeding time slices generated by the real-time density matrix renormalization group method. This optimization can also be applied to classical simulations of quantum circuits. We discuss the time reversal symmetry in the fully optimized MPS.",
            "pub_abstract": "A kind of least action principle is introduced for the discrete time evolution of one-dimensional quantum lattice models. Based on this principle, we obtain an optimal condition for the matrix product states on succeeding time slices generated by the real-time density matrix renormalization group method. This optimization can also be applied to classical simulations of quantum circuits. We discuss the time reversal symmetry in the fully optimized MPS.",
            "title": "Least Action Principle for the Real-Time Density Matrix Renormalization Group",
            "preprint_link": "https://arxiv.org/abs/cond-mat/0612480v2",
        }
        del caller.data["preprint_server"]  # Remove server key before comparison
        self.assertDictEqual(caller.data, correct_data)

    def valid_but_nonexistent_identifier(self):
        caller = ArxivCaller("1613.07611v1")
        self.assertEqual(caller.is_valid, False)


class DOICallerTest(TestCase):
    def test_works_for_physrev_doi(self):
        caller = DOICaller("10.1103/PhysRevB.92.214427")

        correct_data = {
            "title": "Quasi-soliton scattering in quantum spin chains",
            "pages": "214427",
            "author_list": "R. Vlijm, M. Ganahl, D. Fioretto, M. Brockmann, M. Haque, H. G. Evertz, J.-S. Caux",
            "pub_date": "2015-12-18",
            "volume": "92",
            "journal": "Physical Review B",
            "abstract": "",  # Doesn't seem to be available for this DOI
        }
        # Remove crossref_data from caller.data for comparison
        del caller.data["crossref_data"]

        self.assertTrue(caller.is_valid)
        self.assertDictEqual(caller.data, correct_data)

    def test_works_for_scipost_doi(self):
        caller = DOICaller("10.21468/SciPostPhys.2.2.012")
        correct_data = {
            "author_list": "Yannis Brun, Jerome Dubail",
            "pub_date": "2017-04-04",
            "volume": "2",
            "title": "One-particle density matrix of trapped one-dimensional impenetrable bosons from conformal invariance",
            "pages": "012",
            "journal": "SciPost Physics",
            "abstract": "<jats:p>The one-particle density matrix of the one-dimensional\n"
            "Tonks-Girardeau gas with inhomogeneous density profile is calculated,\n"
            "thanks to a recent observation that relates this system to a\n"
            "two-dimensional conformal field theory in curved space. The result is\n"
            "asymptotically exact in the limit of large particle density and small\n"
            "density variation, and holds for arbitrary trapping potentials. In the\n"
            "particular case of a harmonic trap, we recover a formula obtained by\n"
            "Forrester et al. from a different method.</jats:p>",
        }
        self.assertTrue(caller.is_valid)
        # Remove crossref_data from caller.data for comparison
        del caller.data["crossref_data"]
        self.assertDictEqual(caller.data, correct_data)

    def test_valid_but_non_existent_doi(self):
        caller = DOICaller("10.21468/NonExistentJournal.2.2.012")
        self.assertEqual(caller.is_valid, False)
